class _DoNotUpdateType:
    """
    Helper to determine between provided and not provided 'None' parameters in request bodies.

    Usefull in such cases, where it has to be differentiated between a provided 'None'
    and a parameter that wasn't provided, when it is also by default 'None'.

    Examples
    ========

    >>> body = { 'param': None }
    >>> param = body.get('param', DoNotUpdate())
    >>> param
    None

    >>> body = {}
    >>> param = body.get('param', DoNotUpdate())
    >>> param
    DoNotUpdate
    """
    _instance = None

    def __repr__(self):
        """
        Convenience method to pretty print the object name.
        """
        return "DoNotUpdate"

    def __new__(cls):
        """
        New instance creation.

        Prevents several objects of this class from existing. Normally this wouldn't be needed,
        but it's not much code to avoid the edge cases.
        """
        if cls._instance is None:
            cls._instance = super(_DoNotUpdateType, cls).__new__(cls)
        return cls._instance

    def __call__(self):
        """
        Hack to enable usage in annotations. TODO look into better solutions
        """
        return self


DoNotUpdate = _DoNotUpdateType()



import json

from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.url import Url
from b_lambda_layer_common_test.unit.api_gateway import root


def test_FUNC_from_event_WITH_dummy_lambda_event_EXPECT_api_gateway_rest_api_url_extracted():
    """
    Test whether the function can extract parent api url from lambda event.

    :return: No return.
    """
    with open(f'{root}/dummy_event.json', 'r') as file:
        dummy_event = json.loads(file.read())

    assert Url.from_event(dummy_event) == 'https://awsid.execute-api.eu-central-1.amazonaws.com/stage'



from b_lambda_layer_common.package_version import PackageVersion


class Boto3Version(PackageVersion):
    """
    Deprecated ``boto3`` package version identifier.

    Use general ``PackageVersion`` class instead.
    """

    Boto3VersionType = PackageVersion.VersionType



import logging
from typing import Optional, Union, Dict, Any, Iterable, AnyStr

from urllib3 import HTTPResponse

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.api_gateway.url import Url
    from b_lambda_layer_common.util.http_endpoint import HttpEndpoint
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.url import Url
    from b_lambda_layer_common.source.python.b_lambda_layer_common.util.http_endpoint import HttpEndpoint


class NeighbourEndpoint:
    """
    Http caller class that is used to call other endpoints that belong to the same rest api.
    """
    def __init__(
            self,
            lambda_event: Dict[Any, Any],
            relative_endpoint: str,
            method: str,
            body: Optional[Union[Dict[Any, Any], Iterable[Any], AnyStr]] = None,
            headers: Optional[Dict[str, str]] = None
    ) -> None:
        self.__lambda_event = lambda_event
        self.__relative_endpoint = relative_endpoint

        self.__http_endpoint = HttpEndpoint(
            endpoint_url=self.full_endpoint,
            method=method,
            body=body,
            headers=headers
        )

        if self.authorization_token:
            self.__http_endpoint.add_auth_token(self.authorization_token)

    def call(self) -> HTTPResponse:
        return self.__http_endpoint.call_to_response()

    @property
    def http_endpoint(self) -> HttpEndpoint:
        return self.__http_endpoint

    @property
    def parent_api(self) -> str:
        return Url.from_event(self.__lambda_event)

    @property
    def authorization_token(self) -> Optional[str]:
        return self.__lambda_event['headers'].get('authorization')

    @property
    def full_endpoint(self):
        parent_api = self.parent_api.rstrip('/\\').lstrip('/\\')
        relative_endpoint = self.__relative_endpoint.rstrip('/\\').lstrip('/\\')

        return f'{parent_api}/{relative_endpoint}'



from typing import Any, Dict


class ResponseHeaders:
    def __init__(self, headers: Dict[Any, Any]) -> None:
        self.__headers = headers

    @property
    def headers_dict(self):
        return self.__headers

    @staticmethod
    def merge(*headers: 'ResponseHeaders') -> 'ResponseHeaders':
        container = {}

        for head in headers:
            container = {
                **container,
                **head.headers_dict
            }

        return ResponseHeaders(container)

    @staticmethod
    def allow_all_cors_headers() -> 'ResponseHeaders':
        return ResponseHeaders({
            # Required for CORS support to work.
            'Access-Control-Allow-Origin': '*',
            # Required for cookies, authorization headers with HTTPS.
            'Access-Control-Allow-Credentials': True
        })

    @staticmethod
    def json_headers() -> 'ResponseHeaders':
        return ResponseHeaders({
            # Indicate that response is serialized dictionary in JSON format.
            'Content-Type': 'application/json'
        })

    @staticmethod
    def wav_headers() -> 'ResponseHeaders':
        """
        Audio file WAV content type header.
        """
        return ResponseHeaders({
            # Indicate that response is of audio type, wav subtype.
            'Content-Type': 'audio/wav'
        })

    @staticmethod
    def mpeg_headers() -> 'ResponseHeaders':
        """
        Audio file MP3 content type header.
        """
        return ResponseHeaders({
            # Indicate that response is of audio type, mpeg subtype.
            'Content-Type': 'audio/mpeg'
        })

    @staticmethod
    def csv_headers() -> 'ResponseHeaders':
        """
        Text file CSV content type header.
        """
        return ResponseHeaders({
            # Indicate that response is of text type, csv subtype.
            'Content-Type': 'text/csv'
        })



import json
import logging
from typing import Optional, Dict, Any, Union, Iterable, AnyStr

from urllib3 import HTTPResponse

logger = logging.getLogger(__file__)

try:
    from b_lambda_layer_common.util.http_call import HttpCall
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.util.http_call import HttpCall


class HttpEndpoint:
    """
    Class that represents an http endpoint in OOP way. It is a wrapper class.
    """

    def __init__(
            self,
            endpoint_url: str,
            method: str,
            body: Optional[Union[Dict[Any, Any], Iterable[Any], AnyStr]] = None,
            headers: Optional[Dict[str, str]] = None,
            fields: Optional[Any] = None
    ) -> None:
        """
        Constructor.

        :param endpoint_url: Full endpoint url to call.
        :param method: Http request method type.
        :param body: Body to include in request.
        :param headers: Headers to include in request.
        :param fields: Fields can be one of the following:
            # Regular query parameters or request body { key: value } pairs.
            # Multi-value query parameters [ (key: value1), (key: value2) ] for reused key argument.
            # Filetuple, read more it here: https://urllib3.readthedocs.io/en/latest/reference/urllib3.request.html#urllib3.request.RequestMethods.urlopen
        """
        self.__endpoint_url = endpoint_url
        self.__http_method = method
        self.__http_body = body
        self.__headers = headers or {}
        self.__fields = fields

        assert isinstance(self.__endpoint_url, str), 'Expected string.'
        assert self.__http_method in ['GET', 'POST', 'PUT', 'DELETE', 'HEAD', 'OPTIONS'], 'Unsupported method.'

    def add_auth_token(self, auth_token: str) -> 'HttpEndpoint':
        """
        Adds an authorization token to headers and returns itself for chaining.

        :param auth_token: Authorization token.

        :return: Itself.
        """
        self.__headers['Authorization'] = auth_token
        return self

    def call_to_json(self) -> Dict[str, Any]:
        """
        Calls a specified endpoint with specified parameters.

        :return: Http response represented as dictionary.
        """
        data = self._call().data

        if data == b'':
            return {}

        return json.loads(data)

    def call_to_bytes(self) -> bytes:
        """
        Calls a specified endpoint with specified parameters.

        :return: Http response represented as bytes.
        """
        return self._call().data

    def call_to_str(self) -> str:
        """
        Calls a specified endpoint with specified parameters.

        :return: Http response represented as string.
        """
        data = self._call().data
        encodings = ['utf-8', 'ISO-8859-1', 'Windows-1251', 'Windows-1252']

        for en in encodings:
            try:
                return data.decode(encoding=en)
            except (LookupError, UnicodeDecodeError):
                continue

        raise ValueError('Unable to decode data.')

    def call_to_response(self) -> HTTPResponse:
        """
        Calls a specified endpoint with specified parameters.

        :return: Http response.
        """
        return self._call()

    def _call(self) -> HTTPResponse:
        """
        Initiates a call to a specified endpoint with specified parameters.

        :return: Http response object.
        """
        return HttpCall.call(
            method=self.http_method,
            url=self.endpoint_url,
            fields=self.fields,
            headers=self.headers,
            body=self.http_body
        )

    @property
    def endpoint_url(self):
        """
        Url endpoint.

        :return: Url endpoint to call.
        """
        return self.__endpoint_url

    @property
    def http_method(self):
        """
        Endpoint request method type.

        :return: Request method type.
        """
        return self.__http_method

    @property
    def http_body(self) -> Optional[bytes]:
        """
        Endpoint request body represented as bytes.

        :return: Request body.
        """
        if isinstance(self.__http_body, (dict, list, tuple)):
            return json.dumps(self.__http_body).encode('UTF-8')

        if isinstance(self.__http_body, str):
            return self.__http_body.encode('UTF-8')

        if isinstance(self.__http_body, bytes):
            return self.__http_body

        return None

    @property
    def headers(self) -> Dict[str, str]:
        """
        All request headers associated with this endpoint.

        :return: Endpoint request headers.
        """
        headers = {}

        if isinstance(self.__http_body, (dict, list, tuple)):
            headers['Content-Type'] = 'application/json'
        elif isinstance(self.__http_body, str):
            headers['Content-Type'] = 'application/text'
        elif isinstance(self.__http_body, bytes):
            headers['Content-Type'] = 'application/octet-stream'

        return {
            **headers,
            **(self.__headers or {})
        }

    @property
    def fields(self) -> Optional[Any]:
        """
        An appropriate encoding of fields based on the HTTP method used when performing the request.
        An http request using urllib3.request.urlopen will be made using these fields.

        :return: Fields parameter.
        """
        return self.__fields



import os

from b_aws_testing_framework.credentials import Credentials
from b_aws_testing_framework.tools.cdk_testing.cdk_tool_config import CdkToolConfig
from b_aws_testing_framework.tools.cdk_testing.testing_manager import TestingManager

CDK_PATH = f'{os.path.dirname(os.path.abspath(__file__))}'
MANAGER = TestingManager(Credentials(), CdkToolConfig(CDK_PATH, destroy_before_preparing=False))


def pytest_sessionstart(session):
    MANAGER.set_global_prefix()
    MANAGER.prepare_infrastructure()


def pytest_sessionfinish(session, exitstatus):
    MANAGER.destroy_infrastructure()



import logging
from typing import Iterable, List, Tuple, Dict, Any, Union
from boto3 import client

logger = logging.getLogger(__name__)


class Fetcher:
    """
    Fetches parameters from SSM.
    """
    __MAX_PARAMS_SINGLE_FETCH = 10

    def __init__(self, ssm_client: client) -> None:
        """
        Constructor.

        :param ssm_client: Boto3 SSM client instance.
        """
        self.__ssm_client = ssm_client

    def get_parameters(self, names: List[str], with_decryption: bool = True) -> Tuple[Dict[str, Any], List[str]]:
        """
        Fetches parameters from SSM parameter store.

        :param names: Names of the SSM parameters.
        :param with_decryption: Should they be decrypted.

        :return: A tuple of objects:
            1. Dictionary of SSM parameters (items).
            2. List of invalid SSM parameter names.
        """
        items = {}
        invalid_names = []

        for name_batch in Fetcher.__batch(names, Fetcher.__MAX_PARAMS_SINGLE_FETCH):
            response = self.__ssm_client.get_parameters(
                Names=list(name_batch),
                WithDecryption=with_decryption,
            )

            invalid_names.extend(response['InvalidParameters'])

            for item in response['Parameters']:
                item['Value'] = self.__parse_value(item['Value'], item['Type'])
                items[item['Name']] = item

        logger.info(f'Fetched parameters: {items}.')

        return items, invalid_names

    @staticmethod
    def __parse_value(param_value, param_type) -> Union[str, List[str]]:
        """
        Parses value from SSM parameter.

        :param param_value: SSM parameter value.
        :param param_type: SSM parameter type.

        :return: Parsed value.
        """
        if param_type == 'StringList':
            return param_value.split(',')
        return param_value

    @staticmethod
    def __batch(iterable: Iterable, num: int) -> Iterable:
        """
        Turn an iterable into an iterable of batches of size n (or less, for the last one).

        :param iterable: Iterable to make chunks for.
        :param num: How many items per chunk.

        :return: Iterable of chunks.
        """
        length = len(list(iterable))
        for ndx in range(0, length, num):
            yield iterable[ndx:min(ndx + num, length)]



import logging
from abc import ABC
from typing import Optional, Any, Dict

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.exceptions.b_exception import BException
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.b_exception import BException


class HttpException(BException, ABC):
    SERIALIZATION_KEY_CODE = 'code'

    def __init__(self, message: Optional[str] = 'No message'):
        super().__init__(message)

    @staticmethod
    def http_code() -> int:
        raise NotImplementedError()

    def data(self) -> Dict[str, Any]:
        base_data = super().data()
        base_data[self.SERIALIZATION_KEY_CODE] = self.http_code()
        return base_data



import base64
import json
import urllib.parse
from json import JSONDecodeError
from typing import Dict, Any


class Body:
    """
    Class dedicated to parsing API Gateway formed event bodies sent to Lambda functions.
    """

    def __init__(self, event: Dict[str, Any]):
        """
        Constructor.

        :param event: Lambda event.
        """
        self.__body = event['body']
        self.__is_base_64_encoded = event.get('isBase64Encoded', False)

    def from_urlencoded(self) -> Dict[str, Any]:
        """
        Loads a urlencoded body as a dictionary.

        Example:
            in: TaskAttributes=%7B%22from_country%22%3A%22LT%22%7D&TaskPriority=0
            out: {'TaskAttributes': {'from_country': 'LT'}, 'TaskPriority': 0}

        :return: Event body as a dictionary.
        """
        str_body = urllib.parse.unquote(self.decoded())
        parsed_body = {}

        for item in str_body.split('&'):
            # Parameter maxsplit=1 is important here, because sometimes values can be other URLs containing '=' signs.
            # Example:
            #   in: 'UrlAttribute=https://www.example.com/example?param1=one&param2=two'
            #   out with maxsplit=1: ['UrlAttribute', 'https://www.example.com/example?param1=one&param2=two']
            key, value = item.split('=', maxsplit=1)
            try:
                parsed_body[key] = json.loads(value)
            except JSONDecodeError:
                parsed_body[key] = value

        return parsed_body

    def from_json(self) -> Dict[str, Any]:
        """
        Loads a JSON body as a dictionary.

        Example:
            in: "{\"email\":\"email@biomapas.com\",\"group_id\":\"abc1\"}"
            out: {'email': 'email@biomapas.com', 'group_id': 'abc1'}

        :return: Event body as a dictionary.
        """
        return json.loads(self.decoded())

    def decoded(self) -> str:
        """
        Makes sure the body is not Base64 encoded.

        Example:
            in: SGVsbG8gd29ybGQh
            out: Hello world!


        :return: Event body in string format.
        """
        if self.__is_base_64_encoded:
            return base64.b64decode(self.__body).decode()

        return self.__body



from b_lambda_layer_common.boto3_version import Boto3Version


def test_FUNC_from_string_version_WITH_string_version_EXPECT_version_created():
    """
    Test whether a version can be built from a string version.

    :return: No return.
    """
    version = Boto3Version.from_string_version('123')
    assert version.version_string == '123'
    assert version.version_type == Boto3Version.Boto3VersionType.SPECIFIC


def test_FUNC_latest_WITH_no_parameters_EXPECT_version_created():
    """
    Test whether the latest version can be built.

    :return: No return.
    """
    version = Boto3Version.latest()
    assert version.version_string is None
    assert version.version_type == Boto3Version.Boto3VersionType.LATEST


def test_FUNC_dont_install_WITH_no_parameters_EXPECT_version_created():
    """
    Test whether the version can be None.

    :return: No return.
    """
    version = Boto3Version.dont_install()
    assert version.version_string is None
    assert version.version_type == Boto3Version.Boto3VersionType.NONE



import logging
from typing import Optional, Any
from boto3 import client, Session

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.ssm.fetcher import Fetcher
    from b_lambda_layer_common.ssm.refreshable import Refreshable
except ImportError:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.ssm.fetcher import Fetcher
    from b_lambda_layer_common.source.python.b_lambda_layer_common.ssm.refreshable import Refreshable


class SSMParameter(Refreshable):
    """
    SSM Parameter implementation.
    """

    def __init__(
            self,
            param_name: str,
            max_age: int = 0,
            ssm_client: Optional[client] = None
    ) -> None:
        """
        Constructor.

        :param param_name: SSM parameter name for which the value should be fetched.
        :param max_age: Max age of the value until refresh is needed.
        :param ssm_client: Boto3 SSM client (optional).
        """
        self.__ssm_client = ssm_client or Session().client('ssm')
        self.__ssm_fetcher = Fetcher(self.__ssm_client)

        super().__init__(max_age)

        self.__name = param_name
        self.__value: Any = None

    def update_value(self) -> None:
        """
        Force update of the SSM parameter value.

        :return: No return.
        """
        items, invalid_parameters = self.__ssm_fetcher.get_parameters([self.__name])

        if invalid_parameters or self.__name not in items:
            raise ValueError(f"{self.__name} is invalid.")

        self.__value = items[self.__name]['Value']

    @property
    def name(self) -> str:
        """
        Property for the SSM parameter name.

        :return: SSM parameter name.
        """
        return self.__name

    @property
    def value(self) -> Any:
        """
        Property for the SSM parameter value.

        :return: Value of the SSM parameter.
        """
        if self.__value is None:
            logger.info('Cached parameter value is none.')
            self.refresh()

        if self.should_refresh():
            logger.info('Cached parameter value should be refreshed.')
            self.refresh()

        return self.__value



from itertools import cycle


class DummySsmClient:
    def __init__(self):
        self.__dummy_params = cycle([
            {
                'Name': 'TestParameter',
                'Type': 'String',
                'Value': 'StringValue1',
                'Version': 10,
            },
            {
                'Name': 'TestParameter',
                'Type': 'StringList',
                'Value': 'StringValue2',
                'Version': 20,
            },
            {
                'Name': 'TestParameter',
                'Type': 'SecureString',
                'Value': 'StringValue3',
                'Version': 30,
            },
        ])

        self.get_parameters_function_calls = 0

    def get_parameters(self, *args, **kwargs):
        self.get_parameters_function_calls += 1

        return {
            'Parameters': [next(self.__dummy_params)],
            'InvalidParameters': []
        }



import pook
from b_lambda_layer_common.source.python.b_lambda_layer_common.util.http_endpoint import HttpEndpoint


def test_FUNC_call_to_str_WITH_google_endpoint_EXPECT_successfull_call() -> None:
    """
    Checks whether the http endpoint call to str sends request and receives a response.

    :return: No return.
    """
    with pook.use():
        pook.get('https://example.com')
        endpoint = HttpEndpoint(
            endpoint_url='https://example.com',
            method='GET',
        )

        endpoint.call_to_str()


def test_FUNC_add_auth_token_WITH_dummy_value_EXPECT_authorization_token_added() -> None:
    """
    Checks if auth token can be added to the headers.

    :return: No return.
    """
    endpoint = HttpEndpoint(
        endpoint_url='https://google.com',
        method='GET',
    )

    assert 'Authorization' not in endpoint.headers
    endpoint.add_auth_token('test_token')
    assert endpoint.headers['Authorization'] == 'test_token'



from aws_cdk.aws_lambda import Function, Code, Runtime
from aws_cdk.core import Construct
from b_aws_testing_framework.tools.cdk_testing.testing_stack import TestingStack

from b_lambda_layer_common.boto3_version import Boto3Version
from b_lambda_layer_common.layer import Layer
from b_lambda_layer_common.package_version import PackageVersion


class Infrastructure(TestingStack):
    LAMBDA_FUNCTION_NAME_KEY = 'LambdaFunctionName'

    def __init__(self, scope: Construct):
        super().__init__(scope=scope)

        function = Function(
            scope=self,
            id=f'{self.global_prefix()}TestingFunction',
            code=Code.from_inline(
                'from b_lambda_layer_common import api_gateway\n'
                'from b_lambda_layer_common import exceptions\n'
                'from b_lambda_layer_common import ssm\n'
                'from b_lambda_layer_common import util\n'
                'import boto3\n'
                'import botocore\n'
                '\n\n'
                'def handler(*args, **kwargs):\n'
                '    return dict(\n'
                '        Boto3Version=boto3.__version__,\n'
                '        BotocoreVersion=botocore.__version__,\n'
                '    )'
                '\n'
            ),
            handler='index.handler',
            runtime=Runtime.PYTHON_3_6,
            layers=[
                Layer(
                    scope=self,
                    name=f'{self.global_prefix()}TestingLayer',
                    boto3_version=Boto3Version.from_string_version('1.16.35'),
                    botocore_version=PackageVersion.from_string_version('1.19.35')
                )
            ]
        )

        self.add_output(self.LAMBDA_FUNCTION_NAME_KEY, value=function.function_name)

        # Create another function that is not using boto3.
        Function(
            scope=self,
            id=f'{self.global_prefix()}TestingFunction2',
            code=Code.from_inline('def handler(*args, **kwargs): return 200'),
            handler='index.handler',
            runtime=Runtime.PYTHON_3_6,
            layers=[Layer(scope=self, name=f'{self.global_prefix()}TestingLayer2')]
        )



import json
from decimal import Decimal

from b_lambda_layer_common.source.python.b_lambda_layer_common.util.dynamodb_encoder import DynamoDBEncoder


def test_FUNC_default_WITH_decimals_and_sets_EXPECT_successful_serialization() -> None:
    """
    Checks whether serialization works.

    :return: No return.
    """
    d = {
        'key': Decimal(1),
        'key2': {1, 2, 3, 3, 3, 3}
    }

    try:
        json.dumps(d)
    except TypeError:
        pass
    else:
        raise AssertionError('Expected to fail.')

    json.dumps(d, cls=DynamoDBEncoder)



import logging
from functools import wraps

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.api_gateway.response import Response
    from b_lambda_layer_common.api_gateway.response_headers import ResponseHeaders
    from b_lambda_layer_common.exceptions.container.internal_error import InternalError
    from b_lambda_layer_common.exceptions.http_exception import HttpException
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.response import Response
    from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.response_headers import ResponseHeaders
    from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.container.internal_error import InternalError
    from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.http_exception import HttpException


def exception_middleware(func):
    @wraps(func)
    def wrapper_decorator(*args, **kwargs):
        response_headers = ResponseHeaders.merge(
            ResponseHeaders.allow_all_cors_headers(),
            ResponseHeaders.json_headers()
        )

        try:
            return func(*args, **kwargs)
        except HttpException as ex:
            logger.exception('HTTP exception encountered.')

            data = ex.data()

            return Response.json(
                http_status=ex.http_code(),
                headers=response_headers,
                body=data
            )
        except Exception as ex:
            logger.exception('Unexpected exception encountered.')

            error = InternalError(str(ex))

            return Response.json(
                http_status=error.http_code(),
                headers=response_headers,
                body=error.data()
            )

    return wrapper_decorator



from abc import ABC
from typing import Optional, Dict, Any


class BException(Exception, ABC):
    SERIALIZATION_KEY_MESSAGE = 'message'
    SERIALIZATION_KEY_IDENTIFIER = 'identifier'
    SERIALIZATION_KEY_DESCRIPTION = 'description'

    def __init__(self, message: Optional[str] = 'No message'):
        super().__init__(message)

        self.__message = message

    def message(self) -> Optional[str]:
        return self.__message

    @staticmethod
    def identifier() -> str:
        raise NotImplementedError()

    @staticmethod
    def description() -> str:
        raise NotImplementedError()

    def data(self) -> Dict[str, Any]:
        return {
            self.SERIALIZATION_KEY_MESSAGE: self.message(),
            self.SERIALIZATION_KEY_IDENTIFIER: self.identifier(),
            self.SERIALIZATION_KEY_DESCRIPTION: self.description(),
        }



from xml.etree import ElementTree

from b_lambda_layer_common.source.python.b_lambda_layer_common.util.xml_to_dict import XmlToDict


def test_FUNC_convert_WITH_dummy_xml_string_EXPECT_successful_conversion() -> None:
    """
    Checks whether string representing xml can be converted to a dict successfully.

    :return: No return.
    """
    dictionary = XmlToDict('<element></element>').convert()

    assert dictionary == {'element': None}


def test_FUNC_convert_WITH_dummy_xml_document_EXPECT_successful_conversion() -> None:
    """
    Checks whether xml element can be converted to a dict successfully.

    :return: No return.
    """
    xml_element = ElementTree.fromstring(
        '<element>'
        '<list-element>E1</list-element>'
        '<list-element>E2</list-element>'
        '<list-element>E3</list-element>'
        '</element>'
    )

    dictionary = XmlToDict(xml_element).convert()

    assert dictionary == {'element': {'list-element': ['E1', 'E2', 'E3']}}



import logging
from abc import ABC, abstractmethod
from datetime import timedelta, datetime
from typing import Callable, List, Type, Optional
from functools import wraps

logger = logging.getLogger(__name__)


class Refreshable(ABC):
    """
    Abstract class for refreshable objects.
    """

    def __init__(self, max_age: int = 0) -> None:
        """
        Constructor.

        :param max_age: Max age in seconds of how long the refreshable object should live until expiration.
        """
        self.__max_age = max_age
        self.__max_age_delta = timedelta(seconds=max_age)
        self.__last_refresh_time = None

    @abstractmethod
    def update_value(self) -> None:
        """
        Abstract function that each refreshable type object should implement and
        define the logic of fetching and caching the value.

        :return: No return.
        """
        raise NotImplementedError()

    def refresh(self):
        """
        Initiate refresh action which forces to download a new value (update value)
        and updates the last refresh time for determining the next expiry.

        :return: No return.
        """
        logger.info('Refreshing value...')

        # Force update value.
        self.update_value()
        # Keep track of update date for determining the next expiry.
        self.update_refresh_time()

    def should_refresh(self) -> bool:
        """
        Tells whether the value should be (re)updated (refreshed).

        :return: True if should be updated, False otherwise.
        """
        # Never force refresh if no max_age is configured.
        if not self.__max_age:
            return False

        # Always force refresh if values were never fetched.
        if not self.__last_refresh_time:
            return True

        # Force refresh only if max_age seconds have expired.
        return datetime.utcnow() > self.__last_refresh_time + self.__max_age_delta

    def update_refresh_time(self) -> None:
        """
        Update internal refresh reference with current time.

        :return: No return.
        """
        now = datetime.utcnow()
        self.__last_refresh_time = now

    def refresh_on_error(
            self,
            error_classes: Optional[List[Type[Exception]]] = None,
            error_callback: Optional[Callable] = None,
    ) -> Callable:
        """
        Decorator to handle errors and retries.

        :param error_classes: Which errors to handle.
        :param error_callback: Which function to call in case specified errors are raised.

        :return: Decorator function.
        """
        if error_callback and not callable(error_callback):
            raise TypeError("Callback must be callable.")

        def decorator(func):
            """
            Actual func wrapper.
            """
            @wraps(func)
            def wrapped(*args, **kwargs):
                """
                Actual error/retry handling.
                """
                try:
                    return func(*args, **kwargs)
                except tuple(error_classes or [Exception]):
                    logger.exception('Got an error while calling the decorated function.')

                    self.refresh()

                    if error_callback:
                        logger.info(f'Calling callback function {str(error_callback)}...')
                        error_callback()
                    try:
                        logger.info('Calling the decorated function again...')
                        return func(*args, **kwargs)
                    except Exception:
                        logger.exception('After the value was refreshed we still ran into an error.')
                        raise
            return wrapped
        return decorator



import logging
from typing import Dict, Any

logger = logging.getLogger(__name__)


class Url:
    @staticmethod
    def from_event(event: Dict[Any, Any]) -> str:
        """
        Constructs a url of an api gateway resource that called this lambda function.

        :param event: Event passed to a lambda function upon invocation.

        :return: Parent API url.
        """
        try:
            return (
                f'{event["headers"]["X-Forwarded-Proto"]}://'
                f'{event["requestContext"]["domainName"]}/'
                f'{event["requestContext"]["stage"]}'
            )
        except KeyError as ex:
            raise ValueError(f'Can not construct API url from event. Missing keyword: {ex}.')



import json

from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.neighbour_endpoint import NeighbourEndpoint
from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.container.not_reached_error import \
    NotReachedError
from b_lambda_layer_common_test.unit.api_gateway import root

with open(f'{root}/dummy_event.json', 'r') as file:
    dummy_event = json.loads(file.read())


def test_FUNC_call_WITH_dummy_lambda_event_EXPECT_successfull_call_not_reached_error() -> None:
    """
    Test whether the call was successful with a not reached error.

    :return: No return.
    """
    try:
        NeighbourEndpoint(dummy_event, '/something/', 'GET').call()
    except NotReachedError:
        pass
    else:
        raise AssertionError('Expected to fail.')


def test_FUNC_parent_api_WITH_dummy_lambda_event_EXPECT_parent_api_constructed() -> None:
    """
    Test if parent api was successfully constructed from a lambda event.

    :return: No return.
    """
    endpoint = NeighbourEndpoint(dummy_event, '/test', 'GET')
    assert endpoint.parent_api == 'https://awsid.execute-api.eu-central-1.amazonaws.com/stage'


def test_FUNC_authorization_token_WITH_dummy_lambda_event_EXPECT_token_retrieved() -> None:
    """
    Test whether authorization token was successfully retrieved from lambda event.

    :return: No return.
    """
    endpoint = NeighbourEndpoint(dummy_event, '/test', 'GET')
    assert endpoint.authorization_token == 'Bearer token'


def test_FUNC_full_endpoint_WITH_dummy_lambda_event_and_path_part_EXPECT_successfully_constructed() -> None:
    """
    Test whether the full endpoint to call can be constructed.

    :return: No return.
    """
    endpoint = NeighbourEndpoint(dummy_event, '/test/', 'GET')
    assert endpoint.full_endpoint == 'https://awsid.execute-api.eu-central-1.amazonaws.com/stage/test'



from b_lambda_layer_common.source.python.b_lambda_layer_common.ssm.ssm_parameter import SSMParameter
from b_lambda_layer_common_test.unit.ssm.dummy_ssm_client import DummySsmClient


def test_FUNC_refresh_AND_value_WITH_multiple_calls_EXPECT_values_refreshed():
    """
    Test whether the parameter can fetch and refreshed from SSM.

    :return: No return.
    """
    dummy_client = DummySsmClient()

    # Assert #1: No calls to get_parameters were done.
    assert dummy_client.get_parameters_function_calls == 0

    parameter = SSMParameter(
        param_name='TestParameter',
        ssm_client=dummy_client
    )

    # Assert #2: Still no calls to get_parameters.
    assert dummy_client.get_parameters_function_calls == 0

    value1 = parameter.value

    # Assert #3: One call to get_parameters.
    assert dummy_client.get_parameters_function_calls == 1

    value2 = parameter.value

    # Assert #3: Still one call to get_parameters, because value is now cached.
    assert dummy_client.get_parameters_function_calls == 1
    # Assert #4: Both values are the same.
    assert value1 == value2

    parameter.refresh()

    # Assert #5: Second call to get_parameters, because refresh was called.
    assert dummy_client.get_parameters_function_calls == 2

    value3 = parameter.value

    # Assert #6: Still two calls to get_parameters, because value is cached.
    assert dummy_client.get_parameters_function_calls == 2
    # Assert #7: The new refreshed value is different.
    assert (value1 == value2) and (value2 != value3)


def test_FUNC_refresh_on_error_WITH_multiple_decorators_EXPECT_values_refreshed():
    """
    Test whether the decorators work.

    :return: No return.
    """
    dummy_client = DummySsmClient()

    parameter = SSMParameter(
        param_name='TestParameter',
        ssm_client=dummy_client
    )

    outer_decorator_callbacks = 0
    inner_decorator_callbacks = 0

    def callback_function_outer():
        nonlocal outer_decorator_callbacks
        outer_decorator_callbacks += 1

    def callback_function_inner():
        nonlocal inner_decorator_callbacks
        inner_decorator_callbacks += 1

    # Should result in 1st refresh.
    @parameter.refresh_on_error(error_callback=callback_function_outer)
    # Should result in 2nd refresh.
    @parameter.refresh_on_error(error_callback=callback_function_inner)
    def decorated_function():
        nonlocal inner_decorator_callbacks
        nonlocal outer_decorator_callbacks

        # Should result in 3rd refresh.
        print(parameter.value)

        if inner_decorator_callbacks == 0 or outer_decorator_callbacks == 0:
            raise Exception('Lets raise some exceptions!')

    decorated_function()

    assert outer_decorator_callbacks == 1
    assert inner_decorator_callbacks == 1
    assert dummy_client.get_parameters_function_calls == 3



import time
from urllib3.exceptions import HTTPError

import pook
import pytest

from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.container.internal_error import InternalError
from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.container.not_reached_error import NotReachedError
from b_lambda_layer_common.source.python.b_lambda_layer_common.util.http_call import HttpCall


def test_FUNC_call_WITH_responsive_url_EXPECT_call_successful() -> None:
    """
    Check whether the call succeeded.

    :return: No return.
    """
    with pook.use():
        pook.get("https://example.com", status=200)
        status = HttpCall.call('GET', 'https://example.com').status
        assert status == 200

def test_FUNC_call_WITH_not_existing_url_EXPECT_not_reached() -> None:
    """
    Check whether the call failed.

    :return: No return.
    """
    with pook.use():
        mock = pook.get("https://unreachable.com").error(HTTPError()).persist()
        with pytest.raises(NotReachedError):
            HttpCall.call('GET', 'https://unreachable.com')


def test_FUNC_call_WITH_not_existing_endpoint_EXPECT_not_found() -> None:
    """
    Check whether the call failed.

    :return: No return.
    """
    with pook.use():
        pook.get("https://example.com/does-not-exist", status=404)
        with pytest.raises(InternalError, match="Http call failed with status: 404."):
            HttpCall.call('GET', "https://example.com/does-not-exist")

def test_FUNC_call_WITH_network_failures_EXPECT_correct_number_of_retries() -> None:
    """
    Check if the retry logic works correctly.

    :return: No return.
    """
    with pook.use():
        mock = pook.get("https://unreachable.com").error(HTTPError()).times(5)
        with pytest.raises(NotReachedError):
            HttpCall.call('GET', 'https://unreachable.com')
        assert mock.total_matches == 5

def test_FUNC_call_WITH_network_failures_EXPECT_correct_time_waiting() -> None:
    """
    Check if the retry logic works correctly.

    :return: No return.
    """
    with pook.use():
        mock = pook.get("https://unreachable.com").error(HTTPError()).times(5)
        with pytest.raises(NotReachedError):
            start = time.time()
            HttpCall.call('GET', 'https://unreachable.com')
        end = time.time()
        time_taken = end - start
        # HttpCall uses an exponentiation algorithm that should result in about 8 seconds long total wait time.
        assert 7.5 < time_taken < 8.5


from b_lambda_layer_common.source.python.b_lambda_layer_common.ssm.refreshable import Refreshable


def test_FUNC_refresh_on_error_WITH_error_EXPECT_value_refreshed():
    """
    Test whether decorator works.

    :return: No return.
    """

    class DummyParameter(Refreshable):
        def __init__(self):
            super().__init__()

            self.called_counter = 0

        def update_value(self):
            self.called_counter += 1

    dummy_parameter = DummyParameter()

    @dummy_parameter.refresh_on_error()
    def dummy_function():
        raise Exception()

    try:
        dummy_function()
    except:
        pass

    assert dummy_parameter.called_counter == 1



from aws_cdk.core import Stack

from b_lambda_layer_common.layer import Layer


class LayerSingleton:
    """
    Singleton class that creates a Lambda Layer for common usage.
    """
    __instance: 'LayerSingleton' = None
    __lock: bool = True

    @staticmethod
    def is_initialized() -> bool:
        return bool(LayerSingleton.__instance)

    @staticmethod
    def initialize(scope: Stack, name: str):
        if not LayerSingleton.is_initialized():
            LayerSingleton.__lock = False
            LayerSingleton.__instance = LayerSingleton(scope, name)
            LayerSingleton.__lock = True
        else:
            raise Exception('Instance is already initialized.')

    @staticmethod
    def safe_initialize(scope: Stack, name: str):
        if not LayerSingleton.is_initialized():
            LayerSingleton.__lock = False
            LayerSingleton.__instance = LayerSingleton(scope, name)
            LayerSingleton.__lock = True

    @staticmethod
    def get_instance() -> 'LayerSingleton':
        if LayerSingleton.__instance is None:
            raise Exception('Instance is not initialized. Call initialize method.')

        return LayerSingleton.__instance

    def __init__(self, scope: Stack, name: str) -> None:
        """
        Constructor.

        :param scope: A scope in which this resource should be added.
        :param name: Name and an id of the lambda layer.
        """
        if LayerSingleton.__lock:
            raise Exception('You are not allowed to call constructor. Call get_instance method.')

        self.__layer = Layer(scope, name)

    @property
    def layer(self) -> Layer:
        return self.__layer



import base64

from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.response import Response
from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.response_headers import ResponseHeaders


def test_FUNC_message_WITH_test_message_EXPECT_dictionary_returned() -> None:
    """
    Test that the function can create an appropriate dictionary response from text body.

    :return: No return.
    """
    response = Response.message(200, 'test')
    assert response == {'body': '{"message": "test"}', 'isBase64Encoded': False, 'statusCode': 200}


def test_FUNC_json_WITH_test_dictionary_EXPECT_dictionary_returned() -> None:
    """
    Test that the function can create an appropriate dictionary response from dictionary body.

    :return: No return.
    """
    response = Response.json(200, body={'test_key': 'test_value'})
    assert response == {'body': '{"test_key": "test_value"}', 'isBase64Encoded': False, 'statusCode': 200}


def test_FUNC_media_WITH_encoded_file_EXPECT_dictionary_returned() -> None:
    """
    Test that the function will return a byte64 encoded file.

    :return: No return.
    """

    test_value = base64.b64encode(b'test_value')

    response = Response.media(200, ResponseHeaders.wav_headers(), test_value)
    assert response == {'body': test_value, 'isBase64Encoded': True, 'statusCode': 200, 'headers': {'Content-Type': 'audio/wav'}}



from b_lambda_layer_common.source.python.b_lambda_layer_common.events.event_bridge_factory import EventBridgeFactory


def test_FUNC_emit_WITH_valid_event_EXPECT_event_successful() -> None:
    """
    Test the default value for message.

    :return: No return.
    """
    class DummyBoto:
        def put_events(self, *args, **kwargs):
            return {'FailedEntryCount': 0}

    factory = EventBridgeFactory(
        source='Test',
        detail_type='Test',
        detail={}
    )

    factory.emit(boto_client=DummyBoto())



from aws_cdk.core import Stack, App

from b_lambda_layer_common.layer import Layer


def test_FUNC_constructor_WITH_various_parameters_EXPECT_layer_created():
    """
    Test whether the layer can be created.

    :return: No return.
    """
    layer = Layer(Stack(App(), 'Test'), 'Test')
    assert layer is not None



from enum import Enum
from typing import Optional


class PackageVersion:

    class VersionType(Enum):

        NONE = 'NONE'
        SPECIFIC = 'SPECIFIC'
        LATEST = 'LATEST'

    def __init__(self, version: Optional[str] = None, version_type: Optional[VersionType] = None):
        self.__version = version
        self.__version_type = version_type

    @property
    def version_string(self) -> Optional[str]:
        return self.__version

    @property
    def version_type(self) -> VersionType:
        return self.__version_type

    @classmethod
    def from_string_version(cls, version_string: str) -> 'PackageVersion':
        return cls(version=version_string, version_type=cls.VersionType.SPECIFIC)

    @classmethod
    def latest(cls) -> 'PackageVersion':
        return cls(version_type=cls.VersionType.LATEST)

    @classmethod
    def dont_install(cls) -> 'PackageVersion':
        return cls(version_type=cls.VersionType.NONE)



from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.exception_middleware import \
    exception_middleware
from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.container.dependency_error import \
    DependencyError


def test_FUNC_exception_middleware_WITH_function_raises_dependency_error_EXPECT_dependency_error_code_424() -> None:
    """
    Test that exception middleware can actually catch exceptions and return serialized ones.

    :return: No return.
    """
    @exception_middleware
    def dummy_function():
        raise DependencyError()

    data = dummy_function()

    assert data['statusCode'] == 424



from typing import Optional, List

from aws_cdk.aws_lambda import LayerVersion, Runtime, Code
from aws_cdk.core import Stack, AssetHashType, BundlingOptions, BundlingDockerImage

from b_lambda_layer_common.package_version import PackageVersion


class Layer(LayerVersion):
    def __init__(
            self,
            scope: Stack,
            name: str,
            boto3_version: Optional[PackageVersion] = None,
            botocore_version: Optional[PackageVersion] = None
    ):
        boto3_version = boto3_version or PackageVersion.dont_install()
        botocore_version = botocore_version or PackageVersion.dont_install()

        preinstall_command = ['mkdir -p /tmp/asset-output/python/']
        install_command = []
        build_command = []
        bundling_options = None
        asset_hash_type = None

        if boto3_version.version_type == PackageVersion.VersionType.LATEST:
            install_command.append(f'pip install boto3 --upgrade --upgrade-strategy eager -t /tmp/asset-output/python/')
        elif boto3_version.version_type == PackageVersion.VersionType.SPECIFIC:
            install_command.append(f'pip install boto3=={boto3_version.version_string} --upgrade -t /tmp/asset-output/python/')

        if botocore_version.version_type == PackageVersion.VersionType.LATEST:
            install_command.append(f'pip install botocore --upgrade --upgrade-strategy eager -t /tmp/asset-output/python/')
        elif botocore_version.version_type == PackageVersion.VersionType.SPECIFIC:
            install_command.append(f'pip install botocore=={botocore_version.version_string} --upgrade -t /tmp/asset-output/python/')

        # If we have specified an installation command, we must specify build commands.
        if install_command:
            build_command = [
                # Copy.
                'cp -R /tmp/asset-output/. /asset-output/.',
                'cp -R /asset-input/. /asset-output/.',

                # Cleanup.
                'find /asset-output/ -type f -name "*.py[co]" -delete',
                'find /asset-output/ -type d -name "__pycache__" -exec rm -rf {} +',
                'find /asset-output/ -type d -name "*.dist-info" -exec rm -rf {} +',
                'find /asset-output/ -type d -name "*.egg-info" -exec rm -rf {} +',

                # Validation.
                'ls -la /asset-output/python/.',
                'find /asset-output/ -type f -print0 | sort -z | xargs -0 sha1sum | sha1sum'
            ]

        # If build command is specified, bundling options should be specified too.
        if build_command:
            bundling_options = BundlingOptions(
                image=BundlingDockerImage.from_registry('python:3.9'),
                command=[
                    'bash', '-c', ' && '.join(preinstall_command + install_command + build_command)
                ]
            )

        # If bundling options are specified, we must specify asset hash type.
        if bundling_options:
            asset_hash_type = AssetHashType.BUNDLE

        super().__init__(
            scope=scope,
            id=name,
            layer_version_name=name,
            code=Code.from_asset(
                self.get_source_path(),
                asset_hash_type=asset_hash_type,
                bundling=bundling_options
            ),
            compatible_runtimes=self.runtimes()
        )

    @staticmethod
    def get_source_path() -> str:
        from .source import root
        return root

    @staticmethod
    def runtimes() -> Optional[List[Runtime]]:
        return [
            Runtime.PYTHON_3_6,
            Runtime.PYTHON_3_7,
            Runtime.PYTHON_3_8
        ]



import logging
from typing import Optional

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.exceptions.http_exception import HttpException
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.http_exception import HttpException


class ForbiddenError(HttpException):
    def __init__(self, message: Optional[str] = None):
        super().__init__(message)

    @staticmethod
    def http_code() -> int:
        return 403

    @staticmethod
    def identifier() -> str:
        return 'B_FORBIDDEN'

    @staticmethod
    def description() -> str:
        return (
            'The server understood the request but refuses to authorize it. A server that wishes to make public why '
            'the request has been forbidden can describe that reason in the response payload (if any).'
        )



from decimal import Decimal
from json import JSONEncoder
from typing import Any


class DecimalEncoder(JSONEncoder):
    """
    Custom JSONEncoder.

    This encoder checks whether the type of data is Decimal and converts it to float.
    Decimals are usually returned from DynamoDB table if the value is a number.
    """

    def default(self, o: Any):
        if isinstance(o, Decimal):
            return float(o)
        return super(DecimalEncoder, self).default(o)



import logging
from typing import Any

from pynamodb.attributes import MapAttribute

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.util.dynamodb_encoder import DynamoDBEncoder
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.util.dynamodb_encoder import DynamoDBEncoder


class PynamoDBEncoder(DynamoDBEncoder):
    def default(self, o: Any):
        if isinstance(o, MapAttribute):
            return o.as_dict()
        return super(PynamoDBEncoder, self).default(o)



import json

from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.body import Body
from b_lambda_layer_common_test.unit.api_gateway import root


def test_FUNC_from_urlencoded_WITH_urlencoded_body_EXPECT_successfully_parsed() -> None:
    """
    Test that the function can successfully parse url encoded body.

    :return: No return.
    """
    with open(f'{root}/dummy_event.json', 'r') as file:
        dummy_event = json.loads(file.read())

    dummy_event['body'] = 'TaskAttributes=%7B%22from_country%22%3A%22LT%22%7D&TaskPriority=0'

    parsed = Body(dummy_event).from_urlencoded()

    assert parsed == {'TaskAttributes': {'from_country': 'LT'}, 'TaskPriority': 0}


def test_FUNC_from_json_WITH_json_serialized_body_EXPECT_successfuly_parsed() -> None:
    """
    Test that the function can successfully parse json serialized body.

    :return: No return.
    """
    with open(f'{root}/dummy_event.json', 'r') as file:
        dummy_event = json.loads(file.read())

    dummy_event['body'] = "{\"email\":\"email@biomapas.com\",\"group_id\":\"abc1\"}"

    parsed = Body(dummy_event).from_json()

    assert parsed == {'email': 'email@biomapas.com', 'group_id': 'abc1'}



import json

from b_lambda_layer_common.source.python.b_lambda_layer_common.ws_api_gateway.ws_message import WsMessage


def test_FUNC_success_WITH_text_EXPECT_ws_api_gateway_message_formed_and_encoded():
    """
    Test if the message is formed and encoded correctly.

    :return: No return.
    """
    text = 'test_text'

    message = WsMessage.success(text)

    assert json.loads(message.decode('UTF-8')) == {
        'message': text,
        'type': 'SUCCESS'
    }


def test_FUNC_info_WITH_text_EXPECT_ws_api_gateway_message_formed_and_encoded():
    """
    Test if the message is formed and encoded correctly.

    :return: No return.
    """
    text = 'test_text'

    message = WsMessage.info(text)

    assert json.loads(message.decode('UTF-8')) == {
        'message': text,
        'type': 'INFO'
    }


def test_FUNC_warning_WITH_text_EXPECT_ws_api_gateway_message_formed_and_encoded():
    """
    Test if the message is formed and encoded correctly.

    :return: No return.
    """
    text = 'test_text'

    message = WsMessage.warning(text)

    assert json.loads(message.decode('UTF-8')) == {
        'message': text,
        'type': 'WARNING'
    }


def test_FUNC_error_WITH_text_EXPECT_ws_api_gateway_message_formed_and_encoded():
    """
    Test if the message is formed and encoded correctly.

    :return: No return.
    """
    text = 'test_text'

    message = WsMessage.error(text)

    assert json.loads(message.decode('UTF-8')) == {
        'message': text,
        'type': 'ERROR'
    }



import logging
from typing import Dict, Any

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.exceptions.container.unauthorized_error import UnauthorizedError
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.container.unauthorized_error import UnauthorizedError


class CognitoAccessToken:
    """
    Class responsible for parsing Cognito Access Token parameters from API Gateway authorizer parameters.
    """

    def __init__(self, event: Dict[str, Any]) -> None:
        """
        Constructor.

        :param event: Lambda event.
        """
        try:
            self.__auth_token = event['headers']['authorization']
            self.__claims = event['requestContext']['authorizer']['claims']

            assert self.token_use == 'access'
        except (KeyError, AssertionError) as ex:
            raise UnauthorizedError('Missing access token.')

    @property
    def auth_token(self):
        return self.__auth_token

    @property
    def auth_time(self):
        return self.__claims['auth_time']

    @property
    def client_id(self):
        return self.__claims['client_id']

    @property
    def event_id(self):
        return self.__claims['event_id']

    @property
    def exp(self):
        return self.__claims['exp']

    @property
    def iat(self):
        return self.__claims['iat']

    @property
    def iss(self):
        return self.__claims['iss']

    @property
    def jti(self):
        return self.__claims['jti']

    @property
    def sub(self):
        return self.__claims['sub']

    @property
    def token_use(self):
        return self.__claims['token_use']

    @property
    def username(self) -> str:
        return self.__claims['username']



from typing import Any
import logging

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.util.decimal_encoder import DecimalEncoder
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.util.decimal_encoder import DecimalEncoder


class DynamoDBEncoder(DecimalEncoder):
    def default(self, o: Any):
        if isinstance(o, set):
            return list(o)
        return super(DynamoDBEncoder, self).default(o)



import pytest
from aws_cdk.core import Stack, App

from b_lambda_layer_common.layer import Layer
from b_lambda_layer_common.layer_singleton import LayerSingleton


def test_ALL_FUNC_WITH_default_params_EXPECT_singleton_initialized():
    """
    Test whether the layer singleton works.

    :return: No return.
    """
    with pytest.raises(Exception):
        LayerSingleton(Stack(App(), 'Test'), 'Test')

    with pytest.raises(Exception):
        LayerSingleton.get_instance()

    assert LayerSingleton.is_initialized() is False

    LayerSingleton.initialize(Stack(App(), 'Test'), 'Test')

    with pytest.raises(Exception):
        LayerSingleton.initialize(Stack(App(), 'Test'), 'Test')

    LayerSingleton.safe_initialize(Stack(App(), 'Test'), 'Test')

    assert isinstance(LayerSingleton.get_instance().layer, Layer)



from b_lambda_layer_common.source.python.b_lambda_layer_common.util.logging import LoggingManager


def test_FUNC_setup_logging_WITH_nothing_EXPECT_logging_successfully_set_up() -> None:
    """
    Checks whether the call succeeds.

    :return: No return.
    """
    LoggingManager().setup_logging()



import json

from b_aws_testing_framework.credentials import Credentials
from botocore.response import StreamingBody

from b_lambda_layer_common_test.integration.infrastructure import Infrastructure


def test_RESOURCE_lambda_layer_WITH_deployed_lambda_function_EXPECT_execution_successful():
    """
    Test whether the layer provides necessary functionality.

    :return: No return.
    """
    # Create client for lambda service.
    lambda_client = Credentials().boto_session.client('lambda')

    # Invoke specific lambda function.
    response = lambda_client.invoke(
        FunctionName=Infrastructure.get_output(Infrastructure.LAMBDA_FUNCTION_NAME_KEY),
        InvocationType='RequestResponse'
    )

    # Parse the result.
    payload: StreamingBody = response['Payload']
    data = [item.decode() for item in payload.iter_lines()]
    data = json.loads(''.join(data))

    # Assert that the result is as expected.
    assert data['Boto3Version'] == '1.16.35', data
    assert data['BotocoreVersion'] == '1.19.35', data



import logging
from typing import Union, Dict, Any, Type

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.exceptions.container.not_reached_error import NotReachedError
    from b_lambda_layer_common.exceptions.container.dependency_error import DependencyError
    from b_lambda_layer_common.exceptions.container.already_exists_error import AlreadyExistsError
    from b_lambda_layer_common.exceptions.container.bad_request_error import BadRequestError
    from b_lambda_layer_common.exceptions.container.conflict_error import ConflictError
    from b_lambda_layer_common.exceptions.container.failed_rollback_error import FailedRollbackError
    from b_lambda_layer_common.exceptions.container.forbidden_error import ForbiddenError
    from b_lambda_layer_common.exceptions.container.internal_error import InternalError
    from b_lambda_layer_common.exceptions.container.malformed_permission_error import MalformedPermissionError
    from b_lambda_layer_common.exceptions.container.misconfigured_error import MisconfiguredError
    from b_lambda_layer_common.exceptions.container.not_found_error import NotFoundError
    from b_lambda_layer_common.exceptions.container.unauthorized_error import UnauthorizedError
    from b_lambda_layer_common.exceptions.http_exception import HttpException
except ImportError as ex:
    logger.exception(f'Failed import.')
    from .container.not_reached_error import NotReachedError
    from .container.dependency_error import DependencyError
    from .container.already_exists_error import AlreadyExistsError
    from .container.bad_request_error import BadRequestError
    from .container.conflict_error import ConflictError
    from .container.failed_rollback_error import FailedRollbackError
    from .container.forbidden_error import ForbiddenError
    from .container.internal_error import InternalError
    from .container.malformed_permission_error import MalformedPermissionError
    from .container.misconfigured_error import MisconfiguredError
    from .container.not_found_error import NotFoundError
    from .container.unauthorized_error import UnauthorizedError
    from .http_exception import HttpException


class ExceptionMapper:
    __EXCEPTION_MAP: Dict[str, Type[HttpException]] = {
        AlreadyExistsError.identifier(): AlreadyExistsError,
        BadRequestError.identifier(): BadRequestError,
        ConflictError.identifier(): ConflictError,
        DependencyError.identifier(): DependencyError,
        FailedRollbackError.identifier(): FailedRollbackError,
        ForbiddenError.identifier(): ForbiddenError,
        InternalError.identifier(): InternalError,
        MalformedPermissionError.identifier(): MalformedPermissionError,
        MisconfiguredError.identifier(): MisconfiguredError,
        NotFoundError.identifier(): NotFoundError,
        NotReachedError.identifier(): NotReachedError,
        UnauthorizedError.identifier(): UnauthorizedError
    }

    @staticmethod
    def map_and_raise(exception: Union[str, Dict[str, Any], Exception]) -> None:
        logger.info(
            f'Trying to map error...\n'
            f'{exception=}'
        )

        if isinstance(exception, str):
            try:
                raise ExceptionMapper.__EXCEPTION_MAP[exception]()
            except (KeyError, TypeError):
                pass
        elif isinstance(exception, dict):
            try:
                ex_class = ExceptionMapper.__EXCEPTION_MAP[exception['identifier']]
                raise ex_class(exception['message'])
            except (KeyError, TypeError):
                pass
        elif isinstance(exception, Exception):
            raise exception

        raise ValueError('Appropriate exception could not be found.')



import os


class OSParameter:
    def __init__(self, name: str):
        self.__name = name

    @property
    def value(self):
        """
        The value of a given param name in an operating system.
        """
        try:
            value = os.environ[self.__name]
        except KeyError:
            raise KeyError(f'Parameter with a name {self.__name} does not exist.')

        return value



import json
import logging
import time

import urllib3
from urllib3 import HTTPResponse
from urllib3.exceptions import HTTPError

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.exceptions.container.internal_error import InternalError
    from b_lambda_layer_common.exceptions.container.not_reached_error import NotReachedError
    from b_lambda_layer_common.exceptions.exception_mapper import ExceptionMapper
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.container.internal_error import InternalError
    from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.container.not_reached_error import NotReachedError
    from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.exception_mapper import ExceptionMapper


class HttpCall:
    WAIT_EXPONENTIATION_FACTOR = 2
    MAXIMUM_SINGLE_WAIT = 4
    INITIAL_WAIT_TIME = 0.25

    @staticmethod
    def call(method, url, fields=None, headers=None, **urlopen_kw) -> HTTPResponse:
        """
        Wrapper for urllib3 pool manager request method. The reason why we have a wrapper here
        is to be able to raise custom http exceptions that this library provides.

        :param method: Http method e.g. GET.
        :param url: Endpoint url to call.
        :param fields: Dictionary fields to be serialized to JSON.
        :param headers: Headers for the http call.
        :param urlopen_kw: Additional urllib3 request parameters.

        :return: Http response.
        """
        http = urllib3.PoolManager()

        logger.debug(
            f'Calling endpoint...\n'
            f'{method} {url}\n'
            f'{headers=}\n'
            f'{fields=}\n'
            f'{urlopen_kw=}.'
        )

        wait_length = HttpCall.INITIAL_WAIT_TIME
        last_exception = None
        while wait_length <= HttpCall.MAXIMUM_SINGLE_WAIT:
            try:
                try:
                    response: HTTPResponse = http.request(
                        method=method,
                        url=url,
                        fields=fields,
                        headers=headers,
                        **urlopen_kw
                    )
                except HTTPError as ex:
                    raise NotReachedError(str(ex))
                if HttpCall.is_status_repeatable(response.status):
                    try:
                        error_data = json.loads(response.data)
                        ExceptionMapper.map_and_raise(error_data)
                    except ValueError:
                        raise InternalError(f'Http call failed with status: {response.status}.')
            except Exception as ex:
                # Store the last exception thrown to rethrow at a later point.
                last_exception = ex
                time.sleep(wait_length)
                wait_length *= HttpCall.WAIT_EXPONENTIATION_FACTOR
            else:
                # Successful request.
                if 400 <= response.status <= 599:
                    try:
                        error_data = json.loads(response.data)
                        ExceptionMapper.map_and_raise(error_data)
                    except ValueError:
                        raise InternalError(f'Http call failed with status: {response.status}.')
                return response
        # Retry attempts exceeded, raise the last exception.
        raise last_exception

    @staticmethod
    def is_status_repeatable(status:int):
        # 429 is "Too many requests" and we try to repeat on all 5xx errors.
        return (status == 429 or status >=500)



from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.container.already_exists_error import \
    AlreadyExistsError
from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.exception_mapper import ExceptionMapper


def test_FUNC_map_and_raise_WITH_an_already_exists_identifier_EXPECT_appropriate_exception_thrown() -> None:
    """
    Test whether the function can map the exception correctly.

    :return: No return.
    """
    try:
        ExceptionMapper.map_and_raise('B_ALREADY_EXISTS')
    except AlreadyExistsError:
        pass
    else:
        raise AssertionError('Expected to fail.')



from collections import defaultdict
from typing import Dict, Any, Union
from xml.etree import ElementTree
from xml.etree.ElementTree import Element


class XmlToDict:
    """
    Class that is responsible for converting xml to dict.
    """
    def __init__(self, xml_document: Union[str, Element]) -> None:
        """
        Constructor.

        :param xml_document: Xml document.
        """
        if isinstance(xml_document, str):
            self.__xml_document: Element = ElementTree.fromstring(xml_document)
        elif isinstance(xml_document, Element):
            self.__xml_document: Element = xml_document

    def convert(self) -> Dict[Any, Any]:
        """
        Initiates conversion from xml to dict.

        :return: Dictionary representation of an xml document.
        """
        return self.__etree_to_dict(self.__xml_document)

    @staticmethod
    def __etree_to_dict(xml_element: Element) -> Dict[Any, Any]:
        dictionary = {xml_element.tag: {} if xml_element.attrib else None}
        children = list(xml_element)
        if children:
            dd = defaultdict(list)
            for dc in map(XmlToDict.__etree_to_dict, children):
                for k, v in dc.items():
                    dd[k].append(v)
            dictionary = {xml_element.tag: {k: v[0] if len(v) == 1 else v for k, v in dd.items()}}
        if xml_element.attrib:
            dictionary[xml_element.tag].update(('@' + k, v) for k, v in xml_element.attrib.items())
        if xml_element.text:
            text = xml_element.text.strip()
            if children or xml_element.attrib:
                if text:
                    dictionary[xml_element.tag]['#text'] = text
            else:
                dictionary[xml_element.tag] = text
        return dictionary



import json
import logging
from json import JSONEncoder
from typing import Optional, Dict, Any, Type

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.api_gateway.response_headers import ResponseHeaders
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.response_headers import ResponseHeaders


class Response:
    @staticmethod
    def message(
            status: int,
            message: str,
            headers: Optional[ResponseHeaders] = None,
    ) -> Dict[str, Any]:
        return Response.json(http_status=status, headers=headers, body={'message': message})

    @staticmethod
    def json(
            http_status: int,
            headers: Optional[ResponseHeaders] = None,
            body: Optional[Dict[Any, Any]] = None,
            json_encoder: Optional[Type[JSONEncoder]] = None
    ) -> Dict[Any, Any]:
        r = {
            'isBase64Encoded': False,
            'statusCode': http_status
        }

        if headers:
            r['headers'] = headers.headers_dict

        if body:
            r['body'] = json.dumps(body, cls=json_encoder)

        return r

    @staticmethod
    def media(
            http_status: int,
            headers: ResponseHeaders,
            body: bytes
    ) -> Dict[Any, Any]:
        r = {
            'isBase64Encoded': True,
            'statusCode': http_status,
            'headers': headers.headers_dict,
            'body': body
        }

        return r



import logging
from typing import Any, Callable, Set, Union

logger = logging.getLogger(__name__)

# TODO: "Refreshable" class should be moved to a more appropriate location.
try:
    from b_lambda_layer_common.exceptions.container.internal_error import InternalError
    from b_lambda_layer_common.ssm.refreshable import Refreshable
except ImportError:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.container.internal_error import InternalError
    from b_lambda_layer_common.source.python.b_lambda_layer_common.ssm.refreshable import Refreshable


class InMemoryDataCache(Refreshable):
    """
    In-memory data cache implementation.
    """

    def __init__(self, max_age: int = 0):
        """
        Constructor.

        :param max_age: Max age of the cache until reset is needed.
        """
        super().__init__(max_age)

        self.__cache = {}

    def update_value(self) -> None:
        """
        Force reset of the cache.

        :return: No return.
        """
        self.__cache = {}

    def use_cache(self, pointer: Union[str, Set[str]], func: Callable, *args, **kwargs) -> Any:
        """
        Caches function return values.

        :param pointer: A string or a set of strings used to reference cached data.
        :param func: A function whose return value will be cached.
        :param args: Arguments that will be passed to func.

        :key kwargs: Keyword arguments that will be passed to func.

        :return: Cached data.
        """
        if self.should_refresh():
            logger.info('Cached data should be reset.')
            self.refresh()

        if not isinstance(pointer, (str, set)):
            raise InternalError('Type of pointer value must be str or set.')

        return self.__get_data(pointer) or self.__add_data(pointer, func, *args, **kwargs)

    def __get_data(self, pointer: Union[str, Set[str]]) -> Any:
        """
        Retrieves data from cache if possible.

        :param pointer: A string or a set of strings used to reference cached data.

        :return: Data if it's cached, otherwise None.
        """
        if isinstance(pointer, str) and pointer in self.__cache:
            data = self.__cache[pointer]
        elif isinstance(pointer, set) and pointer.issubset(self.__cache):
            data = {key: self.__cache[key] for key in pointer}
        else:
            data = None

        logger.info('Using cached data.' if data else 'Data not in cache.')

        return data

    def __add_data(self, pointer: Union[str, Set[str]], func: Callable, *args, **kwargs) -> Any:
        """
        Adds data to cache.

        :param pointer: A string or a set of strings used to reference cached data.
        :param func: A function whose return value will be cached.
        :param args: Arguments that will be passed to func.

        :key kwargs: Keyword arguments that will be passed to func.

        :return: Cached data.
        """
        logger.info('Adding data to cache...')

        data = func(*args, **kwargs)

        self.__cache.update(
            {pointer: data} if isinstance(pointer, str) else {key: data[key] for key in pointer}
        )

        return data



import json
from enum import Enum


class WsMessageType(Enum):
    SUCCESS = 'SUCCESS'
    INFO = 'INFO'
    WARNING = 'WARNING'
    ERROR = 'ERROR'


class WsMessage:
    @staticmethod
    def success(text: str) -> bytes:
        return WsMessage.__form_message(text, WsMessageType.SUCCESS)

    @staticmethod
    def info(text: str) -> bytes:
        return WsMessage.__form_message(text, WsMessageType.INFO)

    @staticmethod
    def warning(text: str) -> bytes:
        return WsMessage.__form_message(text, WsMessageType.WARNING)

    @staticmethod
    def error(text: str) -> bytes:
        return WsMessage.__form_message(text, WsMessageType.ERROR)

    @staticmethod
    def __form_message(text: str, type: WsMessageType) -> bytes:
        return json.dumps({
            'message': text,
            'type': type.value
        }).encode('UTF-8')



from b_lambda_layer_common.source.python.b_lambda_layer_common.util.do_not_update import DoNotUpdate


def test_FUNC_init_WITH_no_params_EXPECT_class_initiated() -> None:
    """
    Checks whether class initialization works.

    :return: No return.
    """
    DoNotUpdate()


def test_FUNC_init_WITH_no_params_EXPECT_class_is_type() -> None:
    """
    Checks whether class initialization works.

    :return: No return.
    """
    dnu1 = DoNotUpdate()
    dnu2 = DoNotUpdate

    assert dnu1 is dnu2



import os

from b_lambda_layer_common.source.python.b_lambda_layer_common.util.os_parameter import OSParameter


def test_FUNC_value_WITH_existing_os_value_EXPECT_value_returned():
    """
    Test whether the value was retrieved from OS.

    :return: No return.
    """
    os.environ['TestValue-abc-123'] = 'ExampleValue'
    parameter = OSParameter('TestValue-abc-123')

    assert parameter.value == 'ExampleValue'



from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.http_exception import HttpException


def test_FUNC_http_code_WITH_no_implementation_EXPECT_no_implemented_error_thrown() -> None:
    """
    Test whether the exception is thrown whether the function is called.

    :return: No return.
    """
    try:
        HttpException.http_code()
    except NotImplementedError:
        pass
    else:
        raise AssertionError('Expected to fail.')


def test_FUNC_data_WITH_dummy_values_EXPECT_dictionary_returned() -> None:
    """
    Test how the function creates corresponding dictionary.

    :return: No return.
    """
    class Test(HttpException):
        @staticmethod
        def identifier() -> str:
            return 'DUMMY'

        @staticmethod
        def description() -> str:
            return 'Sample'

        @staticmethod
        def http_code() -> int:
            return 500

    assert Test().data() == {
        'message': 'No message',
        'identifier': 'DUMMY',
        'description': 'Sample',
        'code': 500
    }



import json
import logging
from typing import Dict, Any, Optional

import boto3
from botocore.exceptions import ClientError

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.exceptions.container.internal_error import InternalError
    from b_lambda_layer_common.util.os_parameter import OSParameter
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.container.internal_error import InternalError
    from b_lambda_layer_common.source.python.b_lambda_layer_common.util.os_parameter import OSParameter

try:
    EVENT_BUS_NAME = OSParameter('EVENT_BUS_NAME').value
except KeyError:
    EVENT_BUS_NAME = None


class EventBridgeFactory:
    __client: Optional[Any] = None

    def __init__(
            self,
            source: str,
            detail_type: str,
            detail: Dict[Any, Any],
            event_bus_name: Optional[str] = None
    ) -> None:
        self.__source = source
        self.__detail_type = detail_type
        self.__detail = detail
        self.__event_bus_name = event_bus_name or EVENT_BUS_NAME

    def emit(self, fail_on_error: bool = True, boto_client: Optional[Any] = None) -> None:
        if not boto_client and not self.__client:
            self.__client = boto3.client('events')

        boto_client = boto_client or self.__client

        event_entry = dict(
            Source=self.__source,
            DetailType=self.__detail_type,
            EventBusName=self.__event_bus_name,
            Detail=json.dumps(self.__detail),
        )

        try:
            response = boto_client.put_events(Entries=[event_entry])

            if response['FailedEntryCount'] != 0:
                errors = '\n'.join(entry.get('ErrorMessage') for entry in response['Entries'])
                message = f'Failed to emit events: {errors}.'

                if fail_on_error:
                    raise InternalError(message)

                logger.error(message)
        except ClientError as ex:
            message = f'Failed to emit events ({str(ex)}).'

            if fail_on_error:
                raise InternalError(message)

            logger.exception(message)



import json

from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.cognito_access_token import \
    CognitoAccessToken
from b_lambda_layer_common_test.unit.api_gateway import root


def test_FUNC_username_WITH_dummy_lambda_event_EXPECT_successfully_extracted_username():
    """
    Test that the function can extract username from cognito access (auth) token.

    :return: No return.
    """
    with open(f'{root}/dummy_event.json', 'r') as file:
        dummy_event = json.loads(file.read())

    assert CognitoAccessToken(dummy_event).username == 'unique_user_name'



from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.b_exception import BException


def test_FUNC_message_WITH_no_message_EXPECT_default_message() -> None:
    """
    Test the default value for message.

    :return: No return.
    """
    class Test(BException):
        pass

    assert Test().message() == 'No message'


def test_FUNC_identifier_WITH_no_implementation_EXPECT_not_implemented_error() -> None:
    """
    Test whether the exception is thrown when the function is called.

    :return: No return.
    """
    try:
        BException.identifier()
    except NotImplementedError:
        pass
    else:
        raise AssertionError('Expected to fail.')


def test_FUNC_description_WITH_no_implementation_EXPECT_not_implemented_error() -> None:
    """
    Test whether the exception is thrown when the function is called.

    :return: No return.
    """
    try:
        BException.description()
    except NotImplementedError:
        pass
    else:
        raise AssertionError('Expected to fail.')


def test_FUNC_data_WITH_default_dummy_values_EXPECT_dictionary_created() -> None:
    """
    Test how the function creates a dictionary.

    :return: No return.
    """
    class Test(BException):
        @staticmethod
        def identifier() -> str:
            return 'DUMMY'

        @staticmethod
        def description() -> str:
            return 'Sample'

    assert Test().data() == {
        'message': 'No message',
        'identifier': 'DUMMY',
        'description': 'Sample'
    }



import logging
from typing import Optional

logger = logging.getLogger(__name__)

try:
    from b_lambda_layer_common.exceptions.http_exception import HttpException
except ImportError as ex:
    logger.exception(f'Failed import.')
    from b_lambda_layer_common.source.python.b_lambda_layer_common.exceptions.http_exception import HttpException


class ConflictError(HttpException):
    def __init__(self, message: Optional[str] = None):
        super().__init__(message)

    @staticmethod
    def http_code() -> int:
        return 409

    @staticmethod
    def identifier() -> str:
        return 'B_CONFLICT'

    @staticmethod
    def description() -> str:
        return (
            'The request could not be completed due to a conflict with the current state of the target resource. '
            'This code is used in situations where the user might be able to resolve the conflict and resubmit '
            'the request.'
        )



import logging
from typing import Optional


class LoggingManager:
    """
    Class responsible for common logging setup.
    """

    def __init__(
            self,
            *,
            log_level: int = logging.INFO,
            formatter: Optional[logging.Formatter] = None
    ) -> None:
        """
        Constructor.

        :param log_level: Default log level set for logging.
        :param formatter: Default formatter set for logging.
        """

        self.log_level = log_level
        self.formatter = formatter

    def setup_logging(self) -> None:
        """
        Set up the logging with the default parameters.
        """
        try:
            root_logger = logging.getLogger()
            root_logger.setLevel(self.log_level)

            for handler in root_logger.handlers:
                handler.setLevel(self.log_level)
                if self.formatter:
                    handler.setFormatter(self.formatter)

            root_logger.info("Set up logging")
        except:
            # We don't want this to ever fail. If it fails though, it should fail as loud as it can without disruption.
            logging.exception("Failed to set up logging")



from b_lambda_layer_common.source.python.b_lambda_layer_common.api_gateway.response_headers import ResponseHeaders


def test_FUNC_headers_dict_WITH_dummy_headers_EXPECT_same_dummy_headers() -> None:
    """
    Test that the function returns same headers.

    :return: No return.
    """
    headers = ResponseHeaders({'key': 'value'})
    assert headers.headers_dict == {'key': 'value'}


def test_FUNC_merge_WITH_multiple_headers_EXPECT_headers_merged() -> None:
    """
    Test that the function can merge headers.

    :return: No return.
    """
    headers = ResponseHeaders.merge(
        ResponseHeaders({'key1': 'value'}),
        ResponseHeaders({'key2': 'value'}),
        ResponseHeaders({'key3': 'value'})
    )

    assert headers.headers_dict == {
        'key1': 'value',
        'key2': 'value',
        'key3': 'value'
    }


def test_FUNC_allow_all_cors_headers_WITH_nothing_EXPECT_headers_returned() -> None:
    """
    Test that the function returns appropriate headers.

    :return: No return.
    """
    headers = ResponseHeaders.allow_all_cors_headers().headers_dict

    assert headers == {
        'Access-Control-Allow-Origin': '*',
        'Access-Control-Allow-Credentials': True
    }


def test_FUNC_json_headers_WITH_nothing_EXPECT_headers_returned() -> None:
    """
    Test that the function returns appropriate headers.

    :return: No return.
    """
    headers = ResponseHeaders.json_headers().headers_dict

    assert headers == {
        'Content-Type': 'application/json'
    }


def test_FUNC_wav_headers_WITH_nothing_EXPECT_headers_returned() -> None:
    """
    Test that the function returns appropriate headers.

    :return: No return.
    """
    headers = ResponseHeaders.wav_headers().headers_dict

    assert headers == {
        'Content-Type': 'audio/wav'
    }


def test_FUNC_mpeg_headers_WITH_nothing_EXPECT_headers_returned() -> None:
    """
    Test that the function returns appropriate headers.

    :return: No return.
    """
    headers = ResponseHeaders.mpeg_headers().headers_dict

    assert headers == {
        'Content-Type': 'audio/mpeg'
    }
